/*
  Copyright 2006 by Sean Luke
  Licensed under the Academic Free License version 3.0
  See the file "LICENSE" for more information
*/


package fa.util;

import java.util.*;

import ec.*;
import ec.util.*;
import ec.gp.*;
import ec.gp.koza.*;

public class SizefairCrossover extends GPBreedingPipeline
    {
    public static final String P_NUM_TRIES = "tries";
    public static final String P_MAX_SIZE = "max_size";
    public static final int INDS_PRODUCED = 1;
    public static final int NUM_SOURCES = 2;

    /** How the pipeline selects a node from individual 1 */
    public GPNodeSelector nodeselect1;

    /** How many times the pipeline attempts to pick nodes until it gives up. */
    public int numTries;

    public int max_size;

    /** Temporary holding place for parents */
    public GPIndividual parents[];

    public SizefairCrossover() { parents = new GPIndividual[2]; }

    public Parameter defaultBase() { return (new Parameter("xover-sizefair")); }

    public int numSources() { return NUM_SOURCES; }

    public Object clone()
        {
        SizefairCrossover c = (SizefairCrossover)super.clone();

        // deep-cloned stuff
        c.nodeselect1 = (GPNodeSelector)nodeselect1.clone();
        c.parents = parents.clone();

        return c;
        }

    public void setup(final EvolutionState state, final Parameter base)
        {
        super.setup(state,base);

        Parameter def = defaultBase();
        Parameter p = base.push(P_NODESELECTOR).push("0");
        Parameter d = def.push(P_NODESELECTOR).push("0");

        nodeselect1 = (GPNodeSelector)
            (state.parameters.getInstanceForParameter(
                p,d, GPNodeSelector.class));
        nodeselect1.setup(state,p);

        numTries = state.parameters.getInt(base.push(P_NUM_TRIES),
                                           def.push(P_NUM_TRIES),1);
        if (numTries == 0)
            state.output.fatal("GPCrossover Pipeline has an invalid number of tries (it must be >= 1).",base.push(P_NUM_TRIES),def.push(P_NUM_TRIES));

        max_size = state.parameters.getInt(base.push(P_MAX_SIZE),def.push(P_MAX_SIZE),1);
        if (max_size == 0)
            state.output.fatal("GPCrossover Pipeline has an invalid maximum tree size (it must be >= 1).",base.push(P_MAX_SIZE),def.push(P_MAX_SIZE));
        }

    public int typicalIndsProduced()
        {
        return minChildProduction();
        }

/*
    public final boolean verifyPoints(final GPInitializer initializer,
                                      final GPNode inner1, final GPNode inner2)
        {
        // first check to see if inner1 is swap-compatible with inner2
        // on a type basis
	if (!inner1.swapCompatibleWith(initializer, inner2))
		return false;

        // next check to see if inner1 can fit in inner2's spot
        if (inner1.depth() + inner2.atDepth() > maxDepth)
		return false;

        return true;
        }
*/

private static int
tree_size(final GPNode n)
{
	int s = 1;

	for (GPNode p : n.children)
		s += tree_size(p);
	return s;
}

private static class Bins
{
	int mu_lo, mu_hi;
	List<GPNode> lo, hi, same;
	Map<Integer, List<GPNode>> size_classes;
	GPNode root;
	int s0;
	int s_max;
	boolean no_cache;

	Bins(final GPNode root, boolean no_cache)
	{
		this.root = root;
		this.no_cache = no_cache;
		if (!no_cache) {
			size_classes = new HashMap<Integer, List<GPNode>>();
			classify(root);
		}
	}

	private int
	classify(final GPNode n)
	{
		int s = 1;

		for (GPNode p : n.children)
			s += classify(p);

		List<GPNode> v = size_classes.get(s);

		if (v == null) {
			v = new LinkedList<GPNode>();
			size_classes.put(s, v);
		}
		v.add(n);

		return s;
	}

	void
	bin(final int s)
	{
		s0 = s;
		s_max = 1 + (s0 << 1);

		mu_lo = mu_hi = 0;

		lo = new LinkedList<GPNode>();
		hi = new LinkedList<GPNode>();
		same = new LinkedList<GPNode>();

		if (no_cache)
			bin(root);
		else
			for (Map.Entry<Integer, List<GPNode>> p : size_classes.entrySet()) {
				final int s_i = p.getKey();

				if (s_i < s0) {
					lo.addAll(p.getValue());
					mu_lo += s0 - s_i;
				} else if (s_i == s0)
					same.addAll(p.getValue());
				else if (s_i <= s_max) {
					hi.addAll(p.getValue());
					mu_hi += s_i - s0;
				}
			}
	}

	private int
	bin(final GPNode n)
	{
		int s = 1;

		for (GPNode p : n.children)
			s += bin(p);
		if (s < s0) {
			lo.add(n);
			mu_lo += s;
		} else if (s == s0)
			same.add(n);
		else if (s <= s_max) {
			hi.add(n);
			mu_hi += s;
		}
		return s;
	}
}


    public int produce(final int min, 
                       final int max, 
                       final int start,
                       final int subpopulation,
                       final Individual[] inds,
                       final EvolutionState state,
                       final int thr) 

        {
        // how many individuals should we make?
        int n = typicalIndsProduced();
        if (n < min) n = min;
        if (n > max) n = max;
        GPInitializer initializer = ((GPInitializer)state.initializer);
        
	int q = start;
	while (q < n + start) {
	    boolean success = false;

            if (sources[0] == sources[1])
		sources[0].produce(2,2,0,subpopulation,parents,state,thr);
            else {
		sources[0].produce(1,1,0,subpopulation,parents,state,thr);
		sources[1].produce(1,1,1,subpopulation,parents,state,thr);
	    }

	    final int parent_size = tree_size(parents[0].trees[0].child);

	    Bins b = new Bins(parents[1].trees[0].child, numTries == 1);
            
            nodeselect1.reset();	// prepare the nodeselector
            
            GPNode t0 = null;
            GPNode t1 = null;
            
		/*
		 * Langdon's size-fair subtree selection
		 */

	    for (int x = 0; x < numTries; ++x) {
		t0 = nodeselect1.pickNode(state, subpopulation, thr,
					  parents[0], parents[0].trees[0]);

		final int s = tree_size(t0);

		b.bin(s);

		final int n0 = b.same.size();
		final int n_lo = b.lo.size();
		final int n_hi = b.hi.size();

		List<GPNode> pool;

		if (n_lo == 0) {
			if (n0 == 0)
				continue;
			pool = b.same;
		} else if (n_hi == 0) {
			if (n0 == 0 || state.random[thr].nextDouble() > 1.0 / s)
				pool = b.lo;
			else
				pool = b.same;
		} else {
			final double p0 = (n0 == 0) ? 0.0 : 1.0 / s;
			final double w = state.random[thr].nextDouble();

			if (w < p0)
				pool = b.same;
			else {
				double mu_lo = b.mu_lo / (double)n_lo;
				double mu_hi = b.mu_hi / (double)n_hi;
				double p_hi = (1.0 - p0) / (1 + mu_hi / mu_lo);

				if (w < p0 + p_hi)
					pool = b.hi;
				else
					pool = b.lo;
			}
		}
		t1 = pool.get(state.random[thr].nextInt(pool.size()));
		if (parent_size - s + tree_size(t1) <= max_size) {
			success = true;
			break;
		}
	    }

            GPIndividual j1 = parents[0].lightClone();
            j1.trees = new GPTree[1];
	    j1.trees[0] = parents[0].trees[0].lightClone();
	    if (success) {
	        j1.trees[0].child = parents[0].trees[0].child.cloneReplacing(t1, t0);
		j1.evaluated = false;
	    } else
	        j1.trees[0].child = parents[0].trees[0].child.cloneReplacing();
	    j1.trees[0].child.parent = j1.trees[0];
	    j1.trees[0].child.argposition = 0;
            inds[q++] = j1;
	}
        return n;
    }
}
